@@ -1,10 +1,8 @@
 /* Pango
  * basic-fc.c: Basic shaper for FreeType-based backends
  *
- * Copyright (C) 2000, 2007 Red Hat Software
- * Authors:
- *   Owen Taylor <otaylor@redhat.com>
- *   Behdad Esfahbod <behdad@behdad.org>
+ * Copyright (C) 2000 Red Hat Software
+ * Author: Owen Taylor <otaylor@redhat.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -22,14 +20,14 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
 #include <string.h>
 
 #include <glib/gprintf.h>
 #include "pango-engine.h"
 #include "pango-utils.h"
 #include "pangofc-font.h"
-#include "pango-ot.h"
+
+#include "basic-common.h"
 
 /* No extra fields needed */
 typedef PangoEngineShape      BasicEngineFc;
@@ -39,31 +37,25 @@ typedef PangoEngineShapeClass BasicEngineFcClass;
 #define RENDER_TYPE PANGO_RENDER_TYPE_FC
 
 static PangoEngineScriptInfo basic_scripts[] = {
-  /* Listed in OpenType "Standard scripts" standard */
-  { PANGO_SCRIPT_LATIN,    "*" },
-  { PANGO_SCRIPT_CYRILLIC, "*" },
-  { PANGO_SCRIPT_GREEK,    "*" },
   { PANGO_SCRIPT_ARMENIAN, "*" },
-  { PANGO_SCRIPT_GEORGIAN, "*" },
-  { PANGO_SCRIPT_RUNIC,    "*" },
-  { PANGO_SCRIPT_OGHAM,    "*" },
-
-  /* The following are simple and can be shaped easily too */
-
   { PANGO_SCRIPT_BOPOMOFO, "*" },
   { PANGO_SCRIPT_CHEROKEE, "*" },
   { PANGO_SCRIPT_COPTIC,   "*" },
+  { PANGO_SCRIPT_CYRILLIC, "*" },
   { PANGO_SCRIPT_DESERET,  "*" },
   { PANGO_SCRIPT_ETHIOPIC, "*" },
+  { PANGO_SCRIPT_GEORGIAN, "*" },
   { PANGO_SCRIPT_GOTHIC,   "*" },
+  { PANGO_SCRIPT_GREEK,    "*" },
   { PANGO_SCRIPT_HAN,      "*" },
   { PANGO_SCRIPT_HIRAGANA, "*" },
   { PANGO_SCRIPT_KATAKANA, "*" },
+  { PANGO_SCRIPT_LATIN,    "*" },
+  { PANGO_SCRIPT_OGHAM,    "*" },
   { PANGO_SCRIPT_OLD_ITALIC, "*" },
+  { PANGO_SCRIPT_RUNIC,     "*" },
   { PANGO_SCRIPT_CANADIAN_ABORIGINAL, "*" },
   { PANGO_SCRIPT_YI,       "*" },
-
-  /* Unicode-4.0 additions */
   { PANGO_SCRIPT_BRAILLE,  "*" },
   { PANGO_SCRIPT_CYPRIOT,  "*" },
   { PANGO_SCRIPT_LIMBU,    "*" },
@@ -71,15 +63,7 @@ static PangoEngineScriptInfo basic_scripts[] = {
   { PANGO_SCRIPT_SHAVIAN,  "*" },
   { PANGO_SCRIPT_LINEAR_B, "*" },
   { PANGO_SCRIPT_UGARITIC, "*" },
-
-  /* Unicode-4.1 additions */
-  { PANGO_SCRIPT_GLAGOLITIC, "*" },
-
-  /* Unicode-5.0 additions */
-  { PANGO_SCRIPT_CUNEIFORM,  "*" },
-  { PANGO_SCRIPT_PHOENICIAN, "*" },
-
-  /* In fact any script we don't know how to shape can go here */
+    
   { PANGO_SCRIPT_COMMON,   "" }
 };
 
@@ -92,127 +76,153 @@ static PangoEngineInfo script_engines[] = {
   }
 };
 
-static const PangoOTFeatureMap gsub_features[] =
+static void
+swap_range (PangoGlyphString *glyphs,
+	    int               start,
+	    int               end)
 {
-  {"ccmp", PANGO_OT_ALL_GLYPHS},
-  {"locl", PANGO_OT_ALL_GLYPHS},
-  {"liga", PANGO_OT_ALL_GLYPHS},
-  {"clig", PANGO_OT_ALL_GLYPHS}
-};
-
-static const PangoOTFeatureMap gpos_features[] =
-{
-  {"kern", PANGO_OT_ALL_GLYPHS},
-  {"mark", PANGO_OT_ALL_GLYPHS},
-  {"mkmk", PANGO_OT_ALL_GLYPHS}
-};
-
-static const PangoOTFeatureMap vertical_gsub_features[] =
-{
-  {"ccmp", PANGO_OT_ALL_GLYPHS},
-  {"locl", PANGO_OT_ALL_GLYPHS},
-  {"vert", PANGO_OT_ALL_GLYPHS}
-};
-
-static const PangoOTFeatureMap vertical_gpos_features[] =
-{
-  {"vkrn", PANGO_OT_ALL_GLYPHS}
-};
+  int i, j;
+  
+  for (i = start, j = end - 1; i < j; i++, j--)
+    {
+      PangoGlyphInfo glyph_info;
+      gint log_cluster;
+      
+      glyph_info = glyphs->glyphs[i];
+      glyphs->glyphs[i] = glyphs->glyphs[j];
+      glyphs->glyphs[j] = glyph_info;
+      
+      log_cluster = glyphs->log_clusters[i];
+      glyphs->log_clusters[i] = glyphs->log_clusters[j];
+      glyphs->log_clusters[j] = log_cluster;
+    }
+}
 
 static void
+set_glyph (PangoFont        *font,
+	   PangoGlyphString *glyphs,
+	   int               i,
+	   int               offset,
+	   PangoGlyph        glyph)
+{
+  PangoRectangle logical_rect;
+
+  glyphs->glyphs[i].glyph = glyph;
+  
+  glyphs->glyphs[i].geometry.x_offset = 0;
+  glyphs->glyphs[i].geometry.y_offset = 0;
+
+  glyphs->log_clusters[i] = offset;
+
+  pango_font_get_glyph_extents (font, glyphs->glyphs[i].glyph, NULL, &logical_rect);
+  glyphs->glyphs[i].geometry.width = logical_rect.width;
+}
+
+static void 
 basic_engine_shape (PangoEngineShape *engine,
 		    PangoFont        *font,
 		    const char       *text,
 		    gint              length,
-		    const PangoAnalysis *analysis,
+		    PangoAnalysis    *analysis,
 		    PangoGlyphString *glyphs)
 {
-  PangoFcFont *fc_font;
-  FT_Face face;
-  PangoOTRulesetDescription desc;
-  const PangoOTRuleset *ruleset;
-  PangoOTBuffer *buffer;
-  glong n_chars;
-  const char *p;
-  int cluster = 0;
+  PangoFcFont *fc_font = PANGO_FC_FONT (font);
+  int n_chars;
   int i;
+  const char *p;
 
   g_return_if_fail (font != NULL);
   g_return_if_fail (text != NULL);
   g_return_if_fail (length >= 0);
   g_return_if_fail (analysis != NULL);
 
-  fc_font = PANGO_FC_FONT (font);
-  face = pango_fc_font_lock_face (fc_font);
-  if (!face)
-    return;
-
-  buffer = pango_ot_buffer_new (fc_font);
-  pango_ot_buffer_set_rtl (buffer, analysis->level % 2 != 0);
-
   n_chars = g_utf8_strlen (text, length);
   pango_glyph_string_set_size (glyphs, n_chars);
 
+  pango_fc_font_lock_face (fc_font);
+
   p = text;
-  for (i=0; i < n_chars; i++)
+  for (i = 0; i < n_chars; i++)
     {
       gunichar wc;
-      PangoGlyph glyph;
+      gunichar mirrored_ch;
+      PangoGlyph index;
 
       wc = g_utf8_get_char (p);
 
-      if (g_unichar_type (wc) != G_UNICODE_NON_SPACING_MARK)
-	cluster = p - text;
+      if (analysis->level % 2)
+	if (pango_get_mirror_char (wc, &mirrored_ch))
+	  wc = mirrored_ch;
 
+      if (wc == 0xa0)	/* non-break-space */
+	wc = 0x20;
+		
       if (pango_is_zero_width (wc))
-        glyph = PANGO_GLYPH_EMPTY;
-      else
-        {
-	  gunichar c = wc;
-
-	  if (analysis->level % 2)
-	    g_unichar_get_mirror_char (c, &c);
-
-	  glyph = pango_fc_font_get_glyph (fc_font, c);
+	{
+	  set_glyph (font, glyphs, i, p - text, 0);
 	}
+      else
+	{
+	  index = pango_fc_font_get_glyph (fc_font, wc);
 
-      if (!glyph)
-	glyph = PANGO_GET_UNKNOWN_GLYPH (wc);
-
-      pango_ot_buffer_add_glyph (buffer, glyph, 0, cluster);
-
+	  if (!index)
+	    {
+	      set_glyph (font, glyphs, i, p - text,
+			 pango_fc_font_get_unknown_glyph (fc_font, wc));
+	    }
+	  else
+	    {
+	      set_glyph (font, glyphs, i, p - text, index);
+	      
+	      if (g_unichar_type (wc) == G_UNICODE_NON_SPACING_MARK)
+		{
+		  if (i > 0)
+		    {
+		      PangoRectangle logical_rect, ink_rect;
+		      
+		      glyphs->glyphs[i].geometry.width = MAX (glyphs->glyphs[i-1].geometry.width,
+							      glyphs->glyphs[i].geometry.width);
+		      glyphs->glyphs[i-1].geometry.width = 0;
+		      glyphs->log_clusters[i] = glyphs->log_clusters[i-1];
+		      
+		      /* Some heuristics to try to guess how overstrike glyphs are
+		       * done and compensate
+		       */
+		      pango_font_get_glyph_extents (font, glyphs->glyphs[i].glyph, &ink_rect, &logical_rect);
+		      if (logical_rect.width == 0 && ink_rect.x == 0)
+			glyphs->glyphs[i].geometry.x_offset = (glyphs->glyphs[i].geometry.width - ink_rect.width) / 2;
+		    }
+		}
+	    }
+	}
+      
       p = g_utf8_next_char (p);
     }
 
-  desc.script = analysis->script;
-  desc.language = analysis->language;
+  /* Simple bidi support; most bidi languages (Arabic, Hebrew, Syriac)
+   * are in fact handled in other modules.
+   */
+  if (analysis->level % 2)
+    {
+      int start, end;
 
-  if (PANGO_GRAVITY_IS_VERTICAL (analysis->gravity))
-    {
-      desc.n_static_gsub_features = G_N_ELEMENTS (vertical_gsub_features);
-      desc.static_gsub_features = vertical_gsub_features;
-      desc.n_static_gpos_features = G_N_ELEMENTS (vertical_gpos_features);
-      desc.static_gpos_features = vertical_gpos_features;
-    }
-  else
-    {
-      desc.n_static_gsub_features = G_N_ELEMENTS (gsub_features);
-      desc.static_gsub_features = gsub_features;
-      desc.n_static_gpos_features = G_N_ELEMENTS (gpos_features);
-      desc.static_gpos_features = gpos_features;
+      /* Swap all glyphs */
+      swap_range (glyphs, 0, n_chars);
+      
+      /* Now reorder glyphs within each cluster back to LTR */
+      for (start = 0; start < n_chars;)
+	{
+	  end = start;
+	  while (end < n_chars &&
+		 glyphs->log_clusters[end] == glyphs->log_clusters[start])
+	    end++;
+	  
+	  swap_range (glyphs, start, end);
+	  start = end;
+	}
     }
 
-  /* TODO populate other_features from analysis->extra_attrs */
-  desc.n_other_features = 0;
-  desc.other_features = NULL;
-
-  ruleset = pango_ot_ruleset_get_for_description (pango_ot_info_get (face), &desc);
-
-  pango_ot_ruleset_substitute (ruleset, buffer);
-  pango_ot_ruleset_position (ruleset, buffer);
-  pango_ot_buffer_output (buffer, glyphs);
-
-  pango_ot_buffer_destroy (buffer);
+  pango_fc_font_kern_glyphs (fc_font, glyphs);
 
   pango_fc_font_unlock_face (fc_font);
 }
@@ -224,20 +234,20 @@ basic_engine_fc_class_init (PangoEngineShapeClass *class)
 }
 
 PANGO_ENGINE_SHAPE_DEFINE_TYPE (BasicEngineFc, basic_engine_fc,
-				basic_engine_fc_class_init, NULL)
+				basic_engine_fc_class_init, NULL);
 
-void
+void 
 PANGO_MODULE_ENTRY(init) (GTypeModule *module)
 {
   basic_engine_fc_register_type (module);
 }
 
-void
+void 
 PANGO_MODULE_ENTRY(exit) (void)
 {
 }
 
-void
+void 
 PANGO_MODULE_ENTRY(list) (PangoEngineInfo **engines,
 			  int              *n_engines)
 {
