@@ -36,20 +36,22 @@ struct _PangoAttrIterator
 {
   GSList *next_attribute;
   GList *attribute_stack;
-  int start_index;
-  int end_index;
+  guint start_index;
+  guint end_index;
 };
 
-static PangoAttribute *pango_attr_color_new  (const PangoAttrClass *klass,
-                                              guint16               red,
-                                              guint16               green,
-                                              guint16               blue);
-static PangoAttribute *pango_attr_string_new (const PangoAttrClass *klass,
-                                              const char           *str);
-static PangoAttribute *pango_attr_int_new    (const PangoAttrClass *klass,
-                                              int                   value);
-static PangoAttribute *pango_attr_float_new  (const PangoAttrClass *klass,
-                                              double                value);
+static PangoAttribute *pango_attr_color_new         (const PangoAttrClass *klass,
+                                                     guint16               red,
+                                                     guint16               green,
+                                                     guint16               blue);
+static PangoAttribute *pango_attr_string_new        (const PangoAttrClass *klass,
+                                                     const char           *str);
+static PangoAttribute *pango_attr_int_new           (const PangoAttrClass *klass,
+                                                     int                   value);
+static PangoAttribute *pango_attr_float_new         (const PangoAttrClass *klass,
+                                                     double                value);
+static PangoAttribute *pango_attr_size_new_internal (int                   size,
+						     gboolean              absolute);
 
 
 /**
@@ -396,25 +398,88 @@ pango_attr_float_new  (const PangoAttrClass *klass,
   return (PangoAttribute *)result;
 }
 
+static PangoAttribute *
+pango_attr_size_copy (const PangoAttribute *attr)
+{
+  const PangoAttrSize *size_attr = (PangoAttrSize *)attr;
+
+  if (attr->klass->type == PANGO_ATTR_ABSOLUTE_SIZE)
+    return pango_attr_size_new_absolute (size_attr->size);
+  else
+    return pango_attr_size_new (size_attr->size);
+}
+
+static void
+pango_attr_size_destroy (PangoAttribute *attr)
+{
+  g_free (attr);
+}
+
+static gboolean
+pango_attr_size_equal (const PangoAttribute *attr1,
+		       const PangoAttribute *attr2)
+{
+  const PangoAttrSize *size_attr1 = (const PangoAttrSize *)attr1;
+  const PangoAttrSize *size_attr2 = (const PangoAttrSize *)attr2;
+  
+  return size_attr1->size == size_attr2->size;
+}
+
+static PangoAttribute *
+pango_attr_size_new_internal (int size,
+			      gboolean absolute)
+{
+  PangoAttrSize *result;
+  
+  static const PangoAttrClass klass = {
+    PANGO_ATTR_SIZE,
+    pango_attr_size_copy,
+    pango_attr_size_destroy,
+    pango_attr_size_equal
+  };
+  static const PangoAttrClass absolute_klass = {
+    PANGO_ATTR_ABSOLUTE_SIZE,
+    pango_attr_size_copy,
+    pango_attr_size_destroy,
+    pango_attr_size_equal
+  };
+
+  result = g_new (PangoAttrSize, 1);
+  result->attr.klass = absolute ? &absolute_klass : &klass;
+  result->size = size;
+  result->absolute = absolute;
+
+  return (PangoAttribute *)result;
+}
+
 /**
  * pango_attr_size_new:
  * @size: the font size, in #PANGO_SCALE<!-- -->ths of a point.
  * 
- * Create a new font-size attribute.
+ * Create a new font-size attribute in fractional points.
  * 
  * Return value: the new #PangoAttribute.
  **/
 PangoAttribute *
 pango_attr_size_new (int size)
 {
-  static const PangoAttrClass klass = {
-    PANGO_ATTR_SIZE,
-    pango_attr_int_copy,
-    pango_attr_int_destroy,
-    pango_attr_int_equal
-  };
+  return pango_attr_size_new_internal (size, FALSE);
+}
 
-  return pango_attr_int_new (&klass, size);
+/**
+ * pango_attr_size_new_absolute:
+ * @size: the font size, in #PANGO_SCALE<!-- -->ths of a device unit.
+ * 
+ * Create a new font-size attribute in device units.
+ * 
+ * Return value: the new #PangoAttribute.
+ *
+ * Since: 1.8
+ **/
+PangoAttribute *
+pango_attr_size_new_absolute (int size)
+{
+  return pango_attr_size_new_internal (size, TRUE);
 }
 
 /**
@@ -900,13 +965,17 @@ pango_attr_list_new (void)
  * @list: a #PangoAttrList
  * 
  * Increase the reference count of the given attribute list by one.
+ *
+ * Return value: The attribute list passed in (since 1.10)
  **/
-void
+PangoAttrList *
 pango_attr_list_ref (PangoAttrList *list)
 {
-  g_return_if_fail (list != NULL);
+  g_return_val_if_fail (list != NULL, NULL);
 
   list->ref_count++;
+
+  return list;
 }
 
 /**
@@ -985,7 +1054,7 @@ pango_attr_list_insert_internal (PangoAttrList  *list,
 				 gboolean        before)
 {
   GSList *tmp_list, *prev, *link;
-  gint start_index = attr->start_index;
+  guint start_index = attr->start_index;
 
   if (!list->attributes)
     {
@@ -1092,8 +1161,8 @@ pango_attr_list_change (PangoAttrList  *list,
 			PangoAttribute *attr)
 {
   GSList *tmp_list, *prev, *link;
-  gint start_index = attr->start_index;
-  gint end_index = attr->end_index;
+  guint start_index = attr->start_index;
+  guint end_index = attr->end_index;
 
   g_return_if_fail (list != NULL);
 
@@ -1288,9 +1357,9 @@ pango_attr_list_change (PangoAttrList  *list,
  * 
  * This function splices attribute list @other into @list.
  * This operation is equivalent to stretching every attribute
- * applies at position @pos in @list by an amount @len,
+ * that applies at position @pos in @list by an amount @len,
  * and then calling pango_attr_list_change() with a copy
- * of each attributes in @other in sequence (offset in position by @pos).
+ * of each attribute in @other in sequence (offset in position by @pos).
  *
  * This operation proves useful for, for instance, inserting
  * a preedit string in the middle of an edit buffer.
@@ -1302,26 +1371,40 @@ pango_attr_list_splice (PangoAttrList *list,
 			gint           len)
 {
   GSList *tmp_list;
+  guint upos, ulen;
   
   g_return_if_fail (list != NULL);
   g_return_if_fail (other != NULL); 
   g_return_if_fail (pos >= 0);
   g_return_if_fail (len >= 0);
- 
+
+  upos = (guint)pos;
+  ulen = (guint)len;
+
+/* This definition only works when a and b are unsigned; overflow
+ * isn't defined in the C standard for signed integers
+ */
+#define CLAMP_ADD(a,b) (((a) + (b) < (a)) ? G_MAXUINT : (a) + (b))
+  
   tmp_list = list->attributes;
   while (tmp_list)
     {
       PangoAttribute *attr = tmp_list->data;
 
-      if (attr->start_index <= pos)
+      if (attr->start_index <= upos)
 	{
-	  if (attr->end_index > pos)
-	    attr->end_index += len;
+	  if (attr->end_index > upos)
+	    attr->end_index = CLAMP_ADD (attr->end_index, ulen);
 	}
       else
 	{
-	  attr->start_index += len;
-	  attr->end_index += len;
+	  /* This could result in a zero length attribute if it
+	   * gets squashed up against G_MAXUINT, but deleting such
+	   * an element could (in theory) suprise the caller, so
+	   * we don't delete it.
+	   */
+	  attr->start_index = CLAMP_ADD (attr->end_index, ulen);
+	  attr->end_index = CLAMP_ADD (attr->end_index, ulen);
 	}
 
       tmp_list = tmp_list->next;
@@ -1331,13 +1414,17 @@ pango_attr_list_splice (PangoAttrList *list,
   while (tmp_list)
     {
       PangoAttribute *attr = pango_attribute_copy (tmp_list->data);
-      attr->start_index += pos;
-      attr->end_index += pos;
+      attr->start_index = CLAMP_ADD (attr->start_index, upos);
+      attr->end_index = CLAMP_ADD (attr->end_index, upos);
 
+      /* Same as above, the attribute could be squashed to zero-length; here
+       * pango_attr_list_change() will take care of deleting it.
+       */
       pango_attr_list_change (list, attr);
       
       tmp_list = tmp_list->next;
     }
+#undef CLAMP_ADD
 }
 
 /**
@@ -1364,7 +1451,7 @@ pango_attr_list_get_iterator (PangoAttrList  *list)
   iterator->end_index = 0;
 
   if (!pango_attr_iterator_next (iterator))
-    iterator->end_index = G_MAXINT;
+    iterator->end_index = G_MAXUINT;
 
   return iterator;
 }
@@ -1375,7 +1462,11 @@ pango_attr_list_get_iterator (PangoAttrList  *list)
  * @start: location to store the start of the range
  * @end: location to store the end of the range
  * 
- * Get the range of the current segment.
+ * Get the range of the current segment. Note that the
+ * stored return values are signed, not unsigned like
+ * the values in #PangoAttribute. To deal with this API
+ * oversight, stored return values that wouldn't fit into
+ * a signed integer are clamped to %G_MAXINT.
  **/
 void
 pango_attr_iterator_range (PangoAttrIterator *iterator,
@@ -1385,9 +1476,9 @@ pango_attr_iterator_range (PangoAttrIterator *iterator,
   g_return_if_fail (iterator != NULL);
 
   if (start)
-    *start = iterator->start_index;
+    *start = MIN (iterator->start_index, G_MAXINT);
   if (end)
-    *end = iterator->end_index;
+    *end = MIN (iterator->end_index, G_MAXINT);
 }
 
 /**
@@ -1409,7 +1500,7 @@ pango_attr_iterator_next (PangoAttrIterator *iterator)
     return FALSE;
 
   iterator->start_index = iterator->end_index;
-  iterator->end_index = G_MAXINT;
+  iterator->end_index = G_MAXUINT;
   
   tmp_list = iterator->attribute_stack;
   while (tmp_list)
@@ -1620,7 +1711,14 @@ pango_attr_iterator_get_font (PangoAttrIterator     *iterator,
 	  if (!(mask & PANGO_FONT_MASK_SIZE))
 	    {
 	      mask |= PANGO_FONT_MASK_SIZE;
-	      pango_font_description_set_size (desc, ((PangoAttrInt *)attr)->value);
+	      pango_font_description_set_size (desc, ((PangoAttrSize *)attr)->size);
+	    }
+	  break;
+	case PANGO_ATTR_ABSOLUTE_SIZE:
+	  if (!(mask & PANGO_FONT_MASK_SIZE))
+	    {
+	      mask |= PANGO_FONT_MASK_SIZE;
+	      pango_font_description_set_absolute_size (desc, ((PangoAttrSize *)attr)->size);
 	    }
 	  break;
         case PANGO_ATTR_SCALE:
@@ -1741,9 +1839,9 @@ pango_attr_list_filter (PangoAttrList       *list,
 
 /**
  * pango_attr_iterator_get_attrs:
- * @iterator: a #PangAttrIterator
+ * @iterator: a #PangoAttrIterator
  * 
- * Gets a list all attributes a the current position of the
+ * Gets a list of all attributes at the current position of the
  * iterator.
  * 
  * Return value: a list of all attributes for the current range.
