@@ -539,7 +539,7 @@ xml_isspace (char c)
   return c == ' ' || c == '\t' || c == '\n' || c == '\r';
 }
 
-static GMarkupParser pango_markup_parser = {
+static const GMarkupParser pango_markup_parser = {
   start_element_handler,
   end_element_handler,
   text_handler,
@@ -598,6 +598,8 @@ pango_parse_markup (const char                 *markup_text,
    */
   if (attr_list)
     md->attr_list = pango_attr_list_new ();
+  else
+    md->attr_list = NULL;
 
   md->text = g_string_new ("");
   
@@ -765,8 +767,11 @@ big_parse_func      (MarkupData            *md,
   CHECK_NO_ATTRS("big");
 
   /* Grow text one level */
-  tag->scale_level_delta += 1;
-  tag->scale_level += 1;
+  if (tag)
+    {
+      tag->scale_level_delta += 1;
+      tag->scale_level += 1;
+    }
 
   return TRUE;
 }
@@ -814,7 +819,8 @@ parse_absolute_size (OpenTag               *tag,
    */
   factor = scale_factor (level, 1.0);
   add_attribute (tag, pango_attr_scale_new (factor));
-  open_tag_set_absolute_font_scale (tag, factor);
+  if (tag)
+    open_tag_set_absolute_font_scale (tag, factor);
   
   return TRUE;
 }
@@ -849,9 +855,13 @@ span_parse_func     (MarkupData            *md,
   const char *foreground = NULL;
   const char *background = NULL;
   const char *underline = NULL;
+  const char *underline_color = NULL;
   const char *strikethrough = NULL;
+  const char *strikethrough_color = NULL;
   const char *rise = NULL;
+  const char *letter_spacing = NULL;
   const char *lang = NULL;
+  const char *fallback = NULL;
   
   g_markup_parse_context_get_position (context,
                                        &line_number, &char_number);
@@ -911,21 +921,41 @@ span_parse_func     (MarkupData            *md,
           CHECK_DUPLICATE (underline);
           underline = values[i];
         }
+      else if (strcmp (names[i], "underline_color") == 0)
+        {
+          CHECK_DUPLICATE (underline_color);
+          underline_color = values[i];
+        }
       else if (strcmp (names[i], "strikethrough") == 0)
         {
           CHECK_DUPLICATE (strikethrough);
           strikethrough = values[i];
         }
+      else if (strcmp (names[i], "strikethrough_color") == 0)
+        {
+          CHECK_DUPLICATE (strikethrough_color);
+          strikethrough_color = values[i];
+        }
       else if (strcmp (names[i], "rise") == 0)
         {
           CHECK_DUPLICATE (rise);
           rise = values[i];
         }
+      else if (strcmp (names[i], "letter_spacing") == 0)
+        {
+          CHECK_DUPLICATE (letter_spacing);
+          letter_spacing = values[i];
+        }
       else if (strcmp (names[i], "lang") == 0)
         {
           CHECK_DUPLICATE (lang);
           lang = values[i];
         }
+      else if (strcmp (names[i], "fallback") == 0)
+        {
+          CHECK_DUPLICATE (fallback);
+          fallback = values[i];
+        }
       else
         {
           g_set_error (error, G_MARKUP_ERROR,
@@ -948,7 +978,8 @@ span_parse_func     (MarkupData            *md,
       if (parsed)
         {
           add_attribute (tag, pango_attr_font_desc_new (parsed));
-          open_tag_set_absolute_font_size (tag, pango_font_description_get_size (parsed));
+	  if (tag)
+	    open_tag_set_absolute_font_size (tag, pango_font_description_get_size (parsed));
           pango_font_description_free (parsed);
         }
     }
@@ -980,17 +1011,24 @@ span_parse_func     (MarkupData            *md,
             }
 
           add_attribute (tag, pango_attr_size_new (n));
-          open_tag_set_absolute_font_size (tag, n);
+	  if (tag)
+	    open_tag_set_absolute_font_size (tag, n);
         }
       else if (strcmp (size, "smaller") == 0)
         {
-          tag->scale_level_delta -= 1;
-          tag->scale_level -= 1;
+	  if (tag)
+	    {
+	      tag->scale_level_delta -= 1;
+	      tag->scale_level -= 1;
+	    }
         }
       else if (strcmp (size, "larger") == 0)
         {
-          tag->scale_level_delta += 1;
-          tag->scale_level += 1;
+	  if (tag)
+	    {
+	      tag->scale_level_delta += 1;
+	      tag->scale_level += 1;
+	    }
         }
       else if (parse_absolute_size (tag, size))
         ; /* nothing */
@@ -1131,6 +1169,8 @@ span_parse_func     (MarkupData            *md,
         ul = PANGO_UNDERLINE_DOUBLE;
       else if (strcmp (underline, "low") == 0)
         ul = PANGO_UNDERLINE_LOW;
+      else if (strcmp (underline, "error") == 0)
+        ul = PANGO_UNDERLINE_ERROR;
       else if (strcmp (underline, "none") == 0)
         ul = PANGO_UNDERLINE_NONE;
       else
@@ -1149,6 +1189,24 @@ span_parse_func     (MarkupData            *md,
       add_attribute (tag, pango_attr_underline_new (ul));
     }
 
+  if (underline_color)
+    {
+      PangoColor color;
+
+      if (!pango_color_parse (&color, underline_color))
+        {
+          g_set_error (error,
+                       G_MARKUP_ERROR,
+                       G_MARKUP_ERROR_INVALID_CONTENT,
+                       _("Could not parse underline_color color specification "
+                         "'%s' on line %d"),
+                       underline_color, line_number);
+          goto error;
+        }
+
+      add_attribute (tag, pango_attr_underline_color_new (color.red, color.green, color.blue));
+    }
+
   if (strikethrough)
     {
       if (strcmp (strikethrough, "true") == 0)
@@ -1168,12 +1226,49 @@ span_parse_func     (MarkupData            *md,
         }
     }
 
+  if (strikethrough_color)
+    {
+      PangoColor color;
+
+      if (!pango_color_parse (&color, strikethrough_color))
+        {
+          g_set_error (error,
+                       G_MARKUP_ERROR,
+                       G_MARKUP_ERROR_INVALID_CONTENT,
+                       _("Could not parse strikethrough_color color specification "
+                         "'%s' on line %d"),
+                       strikethrough_color, line_number);
+          goto error;
+        }
+
+      add_attribute (tag, pango_attr_strikethrough_color_new (color.red, color.green, color.blue));
+    }
+
+  if (fallback)
+    {
+      if (strcmp (fallback, "true") == 0)
+        add_attribute (tag, pango_attr_fallback_new (TRUE));
+      else if (strcmp (fallback, "false") == 0)
+        add_attribute (tag, pango_attr_fallback_new (FALSE));
+      else
+        {
+          g_set_error (error,
+                       G_MARKUP_ERROR,
+                       G_MARKUP_ERROR_INVALID_CONTENT,
+                       _("'fallback' attribute on <span> tag "
+                         "line %d should have one of the values "
+                         "'true' or 'false': '%s' is not valid"),
+                       line_number, fallback);
+          goto error;
+        }
+    }
+
   if (rise)
     {
       char *end = NULL;
       glong n;
       
-      n = strtol (weight, &end, 10);
+      n = strtol (rise, &end, 10);
 
       if (*end != '\0')
         {
@@ -1190,6 +1285,28 @@ span_parse_func     (MarkupData            *md,
       add_attribute (tag, pango_attr_rise_new (n));
     }
 
+  if (letter_spacing)
+    {
+      char *end = NULL;
+      glong n;
+      
+      n = strtol (letter_spacing, &end, 10);
+
+      if (*end != '\0')
+        {
+          g_set_error (error,
+                       G_MARKUP_ERROR,
+                       G_MARKUP_ERROR_INVALID_CONTENT,
+                       _("Value of 'letter_spacing' attribute on <span> tag "
+                         "on line %d could not be parsed; "
+                         "should be an integer, not '%s'"),
+                       line_number, letter_spacing);
+          goto error;
+        }
+
+      add_attribute (tag, pango_attr_letter_spacing_new (n));
+    }
+
   if (lang)
     {
       add_attribute (tag,
@@ -1257,8 +1374,11 @@ sub_parse_func      (MarkupData            *md,
   CHECK_NO_ATTRS("sub");
 
   /* Shrink font, and set a negative rise */
-  tag->scale_level_delta -= 1;
-  tag->scale_level -= 1;
+  if (tag)
+    {
+      tag->scale_level_delta -= 1;
+      tag->scale_level -= 1;
+    }
 
   add_attribute (tag, pango_attr_rise_new (-SUPERSUB_RISE));
 
@@ -1276,8 +1396,11 @@ sup_parse_func      (MarkupData            *md,
   CHECK_NO_ATTRS("sup");
 
   /* Shrink font, and set a positive rise */
-  tag->scale_level_delta -= 1;
-  tag->scale_level -= 1;
+  if (tag)
+    {
+      tag->scale_level_delta -= 1;
+      tag->scale_level -= 1;
+    }
 
   add_attribute (tag, pango_attr_rise_new (SUPERSUB_RISE));
 
@@ -1295,8 +1418,11 @@ small_parse_func    (MarkupData            *md,
   CHECK_NO_ATTRS("small");
 
   /* Shrink text one level */
-  tag->scale_level_delta -= 1;
-  tag->scale_level -= 1;
+  if (tag)
+    {
+      tag->scale_level_delta -= 1;
+      tag->scale_level -= 1;
+    }
 
   return TRUE;
 }
