@@ -22,6 +22,8 @@
 /* Size of fontset cache */
 #define FONTSET_CACHE_SIZE 16
 
+#include <math.h>
+
 #include "pango-context.h"
 #include "pangofc-fontmap.h"
 #include "pangofc-private.h"
@@ -32,6 +34,8 @@ typedef struct _PangoFcFace         PangoFcFace;
 typedef struct _PangoFcFamily       PangoFcFamily;
 typedef struct _PangoFcPatternSet   PangoFcPatternSet;
 typedef struct _PangoFcFindFuncInfo PangoFcFindFuncInfo;
+typedef struct _FontsetHashKey      FontsetHashKey;
+typedef struct _FontHashKey         FontHashKey;
 
 #define PANGO_FC_TYPE_FAMILY              (pango_fc_family_get_type ())
 #define PANGO_FC_FAMILY(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_FC_TYPE_FAMILY, PangoFcFamily))
@@ -43,17 +47,15 @@ typedef struct _PangoFcFindFuncInfo PangoFcFindFuncInfo;
 
 struct _PangoFcFontMapPrivate
 {
-  /* We have one map from  PangoFontDescription -> PangoXftPatternSet
-   * per language tag.
-   */
-  GList *fontset_hash_list; 
+  GHashTable *fontset_hash; 	/* Maps PangoFontDescription -> PangoFcPatternSet  */
+  
   /* pattern_hash is used to make sure we only store one copy of
    * each identical pattern. (Speeds up lookup).
    */
   GHashTable *pattern_hash; 
   GHashTable *coverage_hash; /* Maps font file name/id -> PangoCoverage */
 
-  GHashTable *fonts; /* Maps XftPattern -> PangoXftFont */
+  GHashTable *font_hash; /* Maps FcPattern -> PangoFcFont */
 
   GQueue *fontset_cache;	/* Recently used fontsets */
 
@@ -61,6 +63,8 @@ struct _PangoFcFontMapPrivate
   PangoFcFamily **families;
   int n_families;		/* -1 == uninitialized */
 
+  double dpi;
+
   /* Decoders */
   GSList *findfuncs;
 
@@ -100,6 +104,8 @@ struct _PangoFcPatternSet
   FcPattern **patterns;
   PangoFontset *fontset;
   GList *cache_link;
+  
+  FontsetHashKey *key;
 };
 
 struct _PangoFcFindFuncInfo
@@ -113,8 +119,6 @@ struct _PangoFcFindFuncInfo
 static GType    pango_fc_family_get_type     (void);
 static GType    pango_fc_face_get_type       (void);
 
-static void          pango_fc_font_map_init          (PangoFcFontMap              *fontmap);
-static void          pango_fc_font_map_class_init    (PangoFontMapClass            *class);
 static void          pango_fc_font_map_finalize      (GObject                      *object);
 static PangoFont *   pango_fc_font_map_load_font     (PangoFontMap                 *fontmap,
 						       PangoContext                 *context,
@@ -127,6 +131,8 @@ static void          pango_fc_font_map_list_families (PangoFontMap
 						       PangoFontFamily            ***families,
 						       int                          *n_families);
 
+static void pango_fc_font_map_cache_fontset (PangoFcFontMap    *fcfontmap,
+					     PangoFcPatternSet *patterns);
 
 static void pango_fc_pattern_set_free      (PangoFcPatternSet *patterns);
 
@@ -137,35 +143,17 @@ static guint    pango_fc_coverage_key_hash  (PangoFcCoverageKey *key);
 static gboolean pango_fc_coverage_key_equal (PangoFcCoverageKey *key1,
 					      PangoFcCoverageKey *key2);
 
-static PangoFontClass *parent_class;
+static guint    font_hash_key_hash  (const FontHashKey *key);
+static gboolean font_hash_key_equal (const FontHashKey *key_a,
+				     const FontHashKey *key_b);
+static void     font_hash_key_free  (FontHashKey       *key);
 
-GType
-pango_fc_font_map_get_type (void)
-{
-  static GType object_type = 0;
+static guint    fontset_hash_key_hash  (const FontsetHashKey *key);
+static gboolean fontset_hash_key_equal (const FontsetHashKey *key_a,
+					const FontsetHashKey *key_b);
+static void     fontset_hash_key_free  (FontsetHashKey       *key);
 
-  if (!object_type)
-    {
-      static const GTypeInfo object_info =
-      {
-        sizeof (PangoFcFontMapClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) pango_fc_font_map_class_init,
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (PangoFcFontMap),
-        0,              /* n_preallocs */
-        (GInstanceInitFunc) pango_fc_font_map_init,
-      };
-      
-      object_type = g_type_register_static (PANGO_TYPE_FONT_MAP,
-                                            "PangoFcFontMap",
-                                            &object_info, 0);
-    }
-  
-  return object_type;
-}
+G_DEFINE_TYPE (PangoFcFontMap, pango_fc_font_map, PANGO_TYPE_FONT_MAP)
 
 static void 
 pango_fc_font_map_init (PangoFcFontMap *fcfontmap)
@@ -189,26 +177,34 @@ pango_fc_font_map_init (PangoFcFontMap *fcfontmap)
 
   priv->n_families = -1;
 
-  priv->fonts = g_hash_table_new ((GHashFunc)g_direct_hash, NULL);
+  priv->font_hash = g_hash_table_new_full ((GHashFunc)font_hash_key_hash,
+					   (GEqualFunc)font_hash_key_equal,
+					   (GDestroyNotify)font_hash_key_free,
+					   NULL);
+  priv->fontset_hash = g_hash_table_new_full ((GHashFunc)fontset_hash_key_hash,
+					      (GEqualFunc)fontset_hash_key_equal,
+					      (GDestroyNotify)fontset_hash_key_free,
+					      (GDestroyNotify)pango_fc_pattern_set_free);
+
   priv->coverage_hash = g_hash_table_new_full ((GHashFunc)pango_fc_coverage_key_hash,
 					       (GEqualFunc)pango_fc_coverage_key_equal,
 					       (GDestroyNotify)g_free,
 					       (GDestroyNotify)pango_coverage_unref);
   priv->fontset_cache = g_queue_new ();
+  priv->dpi = -1;
 }
 
 static void
-pango_fc_font_map_class_init (PangoFontMapClass *class)
+pango_fc_font_map_class_init (PangoFcFontMapClass *class)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (class);
-  
-  parent_class = g_type_class_peek_parent (class);
+  PangoFontMapClass *fontmap_class = PANGO_FONT_MAP_CLASS (class);
   
   object_class->finalize = pango_fc_font_map_finalize;
-  class->load_font = pango_fc_font_map_load_font;
-  class->load_fontset = pango_fc_font_map_load_fontset;
-  class->list_families = pango_fc_font_map_list_families;
-  class->shape_engine_type = PANGO_RENDER_TYPE_FC;
+  fontmap_class->load_font = pango_fc_font_map_load_font;
+  fontmap_class->load_fontset = pango_fc_font_map_load_fontset;
+  fontmap_class->list_families = pango_fc_font_map_list_families;
+  fontmap_class->shape_engine_type = PANGO_RENDER_TYPE_FC;
 
   g_type_class_add_private (object_class, sizeof (PangoFcFontMapPrivate));
 }
@@ -263,83 +259,182 @@ pango_fc_coverage_key_equal (PangoFcCoverageKey *key1,
   return key1->id == key2->id && strcmp (key1->filename, key2->filename) == 0;
 }
 
-typedef struct _FontsetHashListNode FontsetHashListNode;
-
-struct _FontsetHashListNode {
-  GHashTable    *fontset_hash;
+struct _FontsetHashKey {
+  PangoFcFontMap *fontmap;
+  PangoMatrix matrix;
   PangoLanguage *language;
+  PangoFontDescription *desc;
+  int size;			/* scaled via the current DPI */
+  gpointer context_key;
 };
 
-/* Get the description => fontset map for a particular
- * language tag.
+struct _FontHashKey {
+  PangoFcFontMap *fontmap;
+  PangoMatrix matrix;
+  FcPattern *pattern;
+  gpointer context_key;
+};
+
+/* Fowler / Noll / Vo (FNV) Hash (http://www.isthe.com/chongo/tech/comp/fnv/)
+ * 
+ * Not necessarily better than a lot of other hashes, but should be OK, and
+ * well tested with binary data.
  */
-static GHashTable *
-pango_fc_get_fontset_hash (PangoFcFontMap *fcfontmap,
-			   PangoLanguage  *language)
+
+#define FNV_32_PRIME ((guint32)0x01000193)
+#define FNV1_32_INIT ((guint32)0x811c9dc5)
+
+static guint32
+hash_bytes_fnv (unsigned char *buffer,
+		int            len,
+		guint32        hval)
 {
-  PangoFcFontMapPrivate *priv = fcfontmap->priv;
-	
-  /* We treat NULL as a distinct language tag, but
-   * we should actually determine the real language
-   * tag it corresponds to to avoid duplicate entries
-   * in the list.
-   */
-  GList *tmp_list = priv->fontset_hash_list;
-  while (tmp_list)
+  while (len--)
     {
-      FontsetHashListNode *node = tmp_list->data;
-      if (node->language == language)
-	{
-	  if (tmp_list != priv->fontset_hash_list)
-	    {
-	      /* Put the found node at the beginning
-	       */
-	      priv->fontset_hash_list = g_list_remove_link (priv->fontset_hash_list, tmp_list);
-	      priv->fontset_hash_list->prev = tmp_list;
-	      tmp_list->next = priv->fontset_hash_list;
-	      priv->fontset_hash_list = tmp_list;
-	    }
-	  
-	  return node->fontset_hash;
-	}
-      
-      tmp_list = tmp_list->next;
+      hval *= FNV_32_PRIME;
+      hval ^= *buffer++;
     }
+  
+  return hval;
+}
 
-  {
-    FontsetHashListNode *node = g_new (FontsetHashListNode, 1);
-    priv->fontset_hash_list = g_list_prepend (priv->fontset_hash_list, node);
-    
-    node->fontset_hash =
-      g_hash_table_new_full ((GHashFunc)pango_font_description_hash,
-			     (GEqualFunc)pango_font_description_equal,
-			     (GDestroyNotify)pango_font_description_free,
-			     (GDestroyNotify)pango_fc_pattern_set_free);
-    node->language = language;
+static gboolean
+fontset_hash_key_equal (const FontsetHashKey *key_a,
+			const FontsetHashKey *key_b)
+{
+  if (key_a->size == key_b->size &&
+      pango_font_description_equal (key_a->desc, key_b->desc) &&
+      key_a->matrix.xx == key_b->matrix.xx &&
+      key_a->matrix.xy == key_b->matrix.xy &&
+      key_a->matrix.yx == key_b->matrix.yx &&
+      key_a->matrix.yy == key_b->matrix.yy)
+    {
+      if (key_a->context_key)
+	return PANGO_FC_FONT_MAP_GET_CLASS (key_a->fontmap)->context_key_equal (key_a->fontmap,
+										key_a->context_key,
+										key_b->context_key);
+      else
+	return TRUE;
+    }
+  else
+    return FALSE;
+}
 
-    return node->fontset_hash;
-  }
+static guint
+fontset_hash_key_hash (const FontsetHashKey *key)
+{
+    guint32 hash = FNV1_32_INIT;
+
+    /* We do a bytewise hash on the context matrix */
+    hash = hash_bytes_fnv ((unsigned char *)(&key->matrix),
+			   sizeof(double) * 4,
+			   hash);
+
+    if (key->context_key)
+      hash ^= PANGO_FC_FONT_MAP_GET_CLASS (key->fontmap)->context_key_hash (key->fontmap,
+									    key->context_key);
+
+    /* 1237 is just an abitrary prime */
+    return (hash ^
+	    GPOINTER_TO_UINT (key->language) ^
+	    (key->size * 1237) ^
+	    pango_font_description_hash (key->desc));
 }
 
 static void
-pango_fc_clear_pattern_hashes (PangoFcFontMap *fcfontmap)
+fontset_hash_key_free (FontsetHashKey *key)
 {
-  PangoFcFontMapPrivate *priv = fcfontmap->priv;
-  GList *tmp_list;
+  pango_font_description_free (key->desc);
 
-  tmp_list = priv->fontset_hash_list;
-  while (tmp_list)
+  if (key->context_key)
+    PANGO_FC_FONT_MAP_GET_CLASS (key->fontmap)->context_key_free (key->fontmap,
+								  key->context_key);
+  
+  g_free (key);
+}
+
+static FontsetHashKey *
+fontset_hash_key_copy (FontsetHashKey *old)
+{
+  FontsetHashKey *key = g_new (FontsetHashKey, 1);
+  
+  key->fontmap = old->fontmap;
+  key->matrix = old->matrix;
+  key->language = old->language;
+  key->desc = pango_font_description_copy (old->desc);
+  key->size = old->size;
+  if (old->context_key)
+    key->context_key = PANGO_FC_FONT_MAP_GET_CLASS (key->fontmap)->context_key_copy (key->fontmap,
+										     old->context_key);
+  else
+    key->context_key = NULL;
+  
+  return key;
+}
+
+static gboolean
+font_hash_key_equal (const FontHashKey *key_a,
+		     const FontHashKey *key_b)
+{
+  if (key_a->matrix.xx == key_b->matrix.xx &&
+      key_a->matrix.xy == key_b->matrix.xy &&
+      key_a->matrix.yx == key_b->matrix.yx &&
+      key_a->matrix.yy == key_b->matrix.yy &&
+      key_a->pattern == key_b->pattern)
     {
-      FontsetHashListNode *node = tmp_list->data;
-      
-      g_hash_table_destroy (node->fontset_hash);
-      g_free (node);
-      
-      tmp_list = tmp_list->next;
+      if (key_a->context_key)
+	return PANGO_FC_FONT_MAP_GET_CLASS (key_a->fontmap)->context_key_equal (key_a->fontmap,
+										key_a->context_key,
+										key_b->context_key);
+      else
+	return TRUE;
     }
+  else
+    return FALSE;
+}
 
-  g_list_free (priv->fontset_hash_list);
-  priv->fontset_hash_list = NULL;
+static guint
+font_hash_key_hash (const FontHashKey *key)
+{
+    guint32 hash = FNV1_32_INIT;
+
+    /* We do a bytewise hash on the context matrix */
+    hash = hash_bytes_fnv ((unsigned char *)(&key->matrix),
+			   sizeof(double) * 4,
+			   hash);
+
+    if (key->context_key)
+      hash ^= PANGO_FC_FONT_MAP_GET_CLASS (key->fontmap)->context_key_hash (key->fontmap,
+									    key->context_key);
+
+    return (hash ^ GPOINTER_TO_UINT (key->pattern));
+}
+
+static void
+font_hash_key_free (FontHashKey *key)
+{
+  if (key->context_key)
+    PANGO_FC_FONT_MAP_GET_CLASS (key->fontmap)->context_key_free (key->fontmap,
+								  key->context_key);
+  
+  g_free (key);
+}
+
+static FontHashKey *
+font_hash_key_copy (FontHashKey *old)
+{
+  FontHashKey *key = g_new (FontHashKey, 1);
+  
+  key->fontmap = old->fontmap;
+  key->matrix = old->matrix;
+  key->pattern = old->pattern;
+  if (old->context_key)
+    key->context_key = PANGO_FC_FONT_MAP_GET_CLASS (key->fontmap)->context_key_copy (key->fontmap,
+										     old->context_key);
+  else
+    key->context_key = NULL;
+      
+  return key;
 }
 
 /**
@@ -387,8 +482,8 @@ pango_fc_font_map_finalize (GObject *object)
   g_queue_free (priv->fontset_cache);
   g_hash_table_destroy (priv->coverage_hash);
 
-  if (priv->fonts)
-    g_hash_table_destroy (priv->fonts);
+  if (priv->font_hash)
+    g_hash_table_destroy (priv->font_hash);
 
   if (priv->pattern_hash)
     g_hash_table_destroy (priv->pattern_hash);
@@ -404,21 +499,77 @@ pango_fc_font_map_finalize (GObject *object)
       priv->findfuncs = g_slist_delete_link (priv->findfuncs, priv->findfuncs);
     }
 
-  G_OBJECT_CLASS (parent_class)->finalize (object);
+  G_OBJECT_CLASS (pango_fc_font_map_parent_class)->finalize (object);
+}
+
+static void
+get_context_matrix (PangoContext *context,
+		    PangoMatrix *matrix)
+{
+  const PangoMatrix *set_matrix;
+  static const PangoMatrix identity = PANGO_MATRIX_INIT;
+
+  if (context)
+    set_matrix = pango_context_get_matrix (context);
+  else
+    set_matrix = NULL;
+
+  if (set_matrix)
+    *matrix = *set_matrix;
+  else
+    *matrix = identity;
+}
+
+static void
+font_hash_key_for_context (PangoFcFontMap *fcfontmap,
+			   PangoContext   *context,
+			   FontHashKey    *key)
+{
+  key->fontmap = fcfontmap;
+  get_context_matrix (context, &key->matrix);
+  
+  if (PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->context_key_get)
+    key->context_key = (gpointer)PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->context_key_get (fcfontmap, context);
+  else
+    key->context_key = NULL;
 }
 
 /* Add a mapping from xfont->font_pattern to xfont */
 static void
 pango_fc_font_map_add (PangoFcFontMap *fcfontmap,
+		       PangoContext   *context,
 		       PangoFcFont    *fcfont)
 {
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
+  FontHashKey key;
+  FontHashKey *key_copy;
 
   g_assert (fcfont->fontmap == NULL);
-	
-  g_hash_table_insert (priv->fonts,
-		       fcfont->font_pattern,
-		       fcfont);
+
+  fcfont->fontmap = g_object_ref (fcfontmap);
+
+  font_hash_key_for_context (fcfontmap, context, &key);
+  key.pattern = fcfont->font_pattern;
+  
+  key_copy = font_hash_key_copy (&key);
+  fcfont->context_key = key_copy->context_key;
+  fcfont->matrix = key.matrix;
+  
+  g_hash_table_insert (priv->font_hash, key_copy, fcfont);
+}
+
+static PangoFcFont *
+pango_fc_font_map_lookup (PangoFcFontMap *fcfontmap,
+			  PangoContext   *context,
+			  FcPattern      *pattern)
+{
+  PangoFcFontMapPrivate *priv = fcfontmap->priv;
+  FontHashKey key;
+
+  font_hash_key_for_context (fcfontmap, context, &key);
+  key.pattern = pattern;
+
+  return g_hash_table_lookup (priv->font_hash, &key);
 }
 
 /* Remove mapping from xfont->font_pattern to xfont */
@@ -427,10 +578,16 @@ _pango_fc_font_map_remove (PangoFcFontMap *fcfontmap,
 			   PangoFcFont    *fcfont)
 {
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
+  FontHashKey key;
+
+  key.fontmap = fcfontmap;
+  key.matrix = fcfont->matrix;
+  key.pattern = fcfont->font_pattern;
+  key.context_key = fcfont->context_key;
   
-  g_hash_table_remove (priv->fonts,
-		       fcfont->font_pattern);
+  g_hash_table_remove (priv->font_hash, &key);
   fcfont->fontmap = NULL;
+  fcfont->context_key = NULL;
   g_object_unref (fcfontmap);
 }
 
@@ -480,7 +637,7 @@ pango_fc_font_map_list_families (PangoFontMap      *fontmap,
       if (families)
 	*families = NULL;
       if (n_families)
-	n_families = 0;
+	*n_families = 0;
 
       return;
     }
@@ -544,16 +701,36 @@ pango_fc_font_map_list_families (PangoFontMap      *fontmap,
 static int
 pango_fc_convert_weight_to_fc (PangoWeight pango_weight)
 {
-  if (pango_weight < (PANGO_WEIGHT_NORMAL + PANGO_WEIGHT_LIGHT) / 2)
+#ifdef FC_WEIGHT_ULTRABOLD 
+  /* fontconfig 2.1 only had light/medium/demibold/bold/black */
+  if (pango_weight < (PANGO_WEIGHT_ULTRALIGHT + PANGO_WEIGHT_LIGHT) / 2)
+    return FC_WEIGHT_ULTRALIGHT;
+  else if (pango_weight < (PANGO_WEIGHT_LIGHT + PANGO_WEIGHT_NORMAL) / 2)
     return FC_WEIGHT_LIGHT;
-  else if (pango_weight < (PANGO_WEIGHT_NORMAL + 600) / 2)
+  else if (pango_weight < (PANGO_WEIGHT_NORMAL + 500 /* PANGO_WEIGHT_MEDIUM */) / 2)
+    return FC_WEIGHT_NORMAL;
+  else if (pango_weight < (500 /* PANGO_WEIGHT_MEDIUM */ + PANGO_WEIGHT_SEMIBOLD) / 2)
     return FC_WEIGHT_MEDIUM;
-  else if (pango_weight < (600 + PANGO_WEIGHT_BOLD) / 2)
+  else if (pango_weight < (PANGO_WEIGHT_SEMIBOLD + PANGO_WEIGHT_BOLD) / 2)
+    return FC_WEIGHT_DEMIBOLD;
+  else if (pango_weight < (PANGO_WEIGHT_BOLD + PANGO_WEIGHT_ULTRABOLD) / 2)
+    return FC_WEIGHT_BOLD;
+  else if (pango_weight < (PANGO_WEIGHT_ULTRABOLD + PANGO_WEIGHT_HEAVY) / 2)
+    return FC_WEIGHT_ULTRABOLD;
+  else
+    return FC_WEIGHT_BLACK;
+#else  /* fontconfig < 2.2 */
+  if (pango_weight < (PANGO_WEIGHT_LIGHT + PANGO_WEIGHT_NORMAL) / 2)
+    return FC_WEIGHT_LIGHT;
+  else if (pango_weight < (500 /* PANGO_WEIGHT_MEDIUM */ + PANGO_WEIGHT_SEMIBOLD) / 2)
+    return FC_WEIGHT_MEDIUM;
+  else if (pango_weight < (PANGO_WEIGHT_SEMIBOLD + PANGO_WEIGHT_BOLD) / 2)
     return FC_WEIGHT_DEMIBOLD;
   else if (pango_weight < (PANGO_WEIGHT_BOLD + PANGO_WEIGHT_ULTRABOLD) / 2)
     return FC_WEIGHT_BOLD;
   else
     return FC_WEIGHT_BLACK;
+#endif  
 }
 
 static int
@@ -603,12 +780,14 @@ pango_fc_convert_width_to_fc (PangoStretch pango_stretch)
 #endif
 
 static FcPattern *
-pango_fc_make_pattern (const PangoFontDescription *description)
+pango_fc_make_pattern (const  PangoFontDescription *description,
+		       PangoLanguage               *language,
+		       double                       pixel_size,
+		       double                       dpi)
 {
   FcPattern *pattern;
   int slant;
   int weight;
-  double size;
   char **families;
   int i;
 #ifdef FC_WIDTH
@@ -621,15 +800,22 @@ pango_fc_make_pattern (const PangoFontDescription *description)
   width = pango_fc_convert_width_to_fc (pango_font_description_get_stretch (description));
 #endif
 
-  size = (double) pango_font_description_get_size (description) / PANGO_SCALE;
-  
-  pattern = FcPatternBuild (0,
+  /* The reason for passing in FC_SIZE as well as FC_PIXEL_SIZE is
+   * to work around a bug in libgnomeprint where it doesn't look
+   * for FC_PIXEL_SIZE. See http://bugzilla.gnome.org/show_bug.cgi?id=169020
+   *
+   * Putting FC_SIZE in here slightly reduces the efficiency
+   * of caching of patterns and fonts when working with multiple different
+   * dpi values.
+   */
+  pattern = FcPatternBuild (NULL,
 			    FC_WEIGHT, FcTypeInteger, weight,
 			    FC_SLANT,  FcTypeInteger, slant,
 #ifdef FC_WIDTH
 			    FC_WIDTH,  FcTypeInteger, width,
 #endif
-			    FC_SIZE,   FcTypeDouble,  size,
+ 			    FC_SIZE,  FcTypeDouble,  pixel_size * (72. / dpi),
+ 			    FC_PIXEL_SIZE,  FcTypeDouble,  pixel_size,
 			    NULL);
 
   families = g_strsplit (pango_font_description_get_family (description), ",", -1);
@@ -639,14 +825,19 @@ pango_fc_make_pattern (const PangoFontDescription *description)
 
   g_strfreev (families);
 
+  if (language)
+    FcPatternAddString (pattern, FC_LANG, (FcChar8 *) pango_language_to_string (language));
+  
   return pattern;
 }
 
 static PangoFont *
-pango_fc_font_map_new_font (PangoFontMap  *fontmap,
-			    PangoMatrix   *pango_matrix,
-			    FcPattern     *match)
+pango_fc_font_map_new_font (PangoFontMap               *fontmap,
+			    PangoContext               *context,
+			    const PangoFontDescription *description,
+			    FcPattern                  *match)
 {
+  PangoFcFontMapClass *class;
   PangoFcFontMap *fcfontmap = (PangoFcFontMap *)fontmap;
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
   FcPattern *pattern;
@@ -662,41 +853,52 @@ pango_fc_font_map_new_font (PangoFontMap  *fontmap,
   if (priv->closed)
     return NULL;
   
-  /* Look up cache */
-  if (!pango_matrix)
-    {
-      fcfont = g_hash_table_lookup (priv->fonts, match);
+  fcfont = pango_fc_font_map_lookup (fcfontmap, context, match);
+  if (fcfont)
+    return g_object_ref (fcfont);
   
-      if (fcfont)
-	return g_object_ref (fcfont);
-    }
+  class = PANGO_FC_FONT_MAP_GET_CLASS (fontmap);
 
-  if (pango_matrix)
+  if (class->create_font)
     {
-      FcMatrix fc_matrix;
-
-      /* FontConfig has the Y axis pointing up, Pango, down.
-       */
-      fc_matrix.xx = pango_matrix->xx;
-      fc_matrix.xy = - pango_matrix->xy;
-      fc_matrix.yx = - pango_matrix->yx;
-      fc_matrix.yy = pango_matrix->yy;
-      
-      pattern = FcPatternDuplicate (match);
-      FcPatternAddMatrix (pattern, FC_MATRIX, &fc_matrix);
+      fcfont = PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->create_font (fcfontmap,
+								   context,
+								   description,
+								   match);
     }
   else
-    pattern = match;
+    {
+      const PangoMatrix *pango_matrix;
 
-  fcfont = PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->new_font (fcfontmap, pattern);
+      if (context)
+	pango_matrix = pango_context_get_matrix (context);
+      else
+	pango_matrix = NULL;
 
-  if (!pango_matrix)
-    pango_fc_font_map_add (fcfontmap, fcfont);
+      if (pango_matrix)
+	{
+	  FcMatrix fc_matrix;
+	  
+	  /* FontConfig has the Y axis pointing up, Pango, down.
+	   */
+	  fc_matrix.xx = pango_matrix->xx;
+	  fc_matrix.xy = - pango_matrix->xy;
+	  fc_matrix.yx = - pango_matrix->yx;
+	  fc_matrix.yy = pango_matrix->yy;
+	  
+	  pattern = FcPatternDuplicate (match);
+	  FcPatternAddMatrix (pattern, FC_MATRIX, &fc_matrix);
+	}
+      else
+	pattern = match;
+      
+      fcfont = PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->new_font (fcfontmap, pattern);
 
-  if (pango_matrix)
-    FcPatternDestroy (pattern);
+      if (pango_matrix)
+	FcPatternDestroy (pattern);
+    }
 
-  fcfont->fontmap = g_object_ref (fcfontmap);
+  pango_fc_font_map_add (fcfontmap, context, fcfont);
 
   /*
    * Give any custom decoders a crack at this font now that it's been
@@ -748,41 +950,142 @@ uniquify_pattern (PangoFcFontMap *fcfontmap,
 
 static void
 pango_fc_default_substitute (PangoFcFontMap    *fontmap,
-			     FcPattern          *pattern)
+			     PangoContext      *context,
+			     FcPattern         *pattern)
 {
-  if (PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->default_substitute)
+  if (PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->context_substitute)
+    PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->context_substitute (fontmap, context, pattern);
+  else if (PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->default_substitute)
     PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->default_substitute (fontmap, pattern);
 }
 
+static gdouble
+pango_fc_font_map_get_resolution (PangoFcFontMap *fcfontmap,
+				  PangoContext   *context)
+{
+  if (PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->get_resolution)
+    return PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->get_resolution (fcfontmap, context);
+  
+  if (fcfontmap->priv->dpi < 0)
+    {
+      FcResult result = FcResultNoMatch;
+      FcPattern *tmp = FcPatternBuild (NULL,
+				       FC_FAMILY, FcTypeString, "Sans",
+				       FC_SIZE,   FcTypeDouble, 10.,
+				       NULL);
+      if (tmp)
+	{
+	  pango_fc_default_substitute (fcfontmap, NULL, tmp);
+	  result = FcPatternGetDouble (tmp, FC_DPI, 0, &fcfontmap->priv->dpi);
+	  FcPatternDestroy (tmp);
+	}
+      
+      if (result != FcResultMatch)
+	{
+	  g_warning ("Error getting DPI from fontconfig, using 72.0");
+	  fcfontmap->priv->dpi = 72.0;
+	}
+    }
+
+  return fcfontmap->priv->dpi;
+}
+
+static int
+get_unscaled_size (PangoFcFontMap             *fcfontmap,
+		   PangoContext               *context,
+		   const PangoFontDescription *desc)
+{
+  int size = pango_font_description_get_size (desc);
+  
+  if (pango_font_description_get_size_is_absolute (desc))
+    return size;
+  else
+    {
+      double dpi = pango_fc_font_map_get_resolution (fcfontmap, context);
+
+      return (int)(0.5 + size * dpi / 72.);
+    }
+}
+
+/*
+ * Based on cairo-matrix.c:_cairo_matrix_compute_scale_factors()
+ *
+ * Copyright 2005, Keith Packard
+ */
+static int
+get_scaled_size (FontsetHashKey *key)
+{
+  PangoMatrix *matrix = &key->matrix;
+  double det = matrix->xx * matrix->yy - matrix->yx * matrix->xy;
+
+  if (det == 0)
+    {
+      return 0.0;
+    }
+  else
+    {
+      double x = matrix->xx;
+      double y = matrix->yx;
+      double major, minor;
+
+      major = sqrt (x*x + y*y);
+      
+      /*
+       * ignore mirroring
+       */
+      if (det < 0)
+	det = - det;
+      
+      if (major)
+	minor = det / major;
+      else 
+	minor = 0.0;
+
+      return minor * (key->size / 1024.);
+    }
+}
+
 static PangoFcPatternSet *
 pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
-				 PangoContext               *context,
-				 const PangoFontDescription *desc,
-				 PangoLanguage              *language)
+				PangoContext               *context,
+				const PangoFontDescription *desc,
+				PangoLanguage              *language)
 {
   PangoFcFontMap *fcfontmap = (PangoFcFontMap *)fontmap;
+  PangoFcFontMapPrivate *priv = fcfontmap->priv;
   FcPattern *pattern, *font_pattern;
   FcResult res;
   int f;
   PangoFcPatternSet *patterns;
   FcFontSet *font_patterns;
-  GHashTable *fontset_hash;
+  FontsetHashKey key;
 
   if (!language && context)
     language = pango_context_get_language (context);
+
+  key.fontmap = fcfontmap;
+  get_context_matrix (context, &key.matrix);
+  key.language = language;
+  key.desc = pango_font_description_copy_static (desc);
+  pango_font_description_unset_fields (key.desc, PANGO_FONT_MASK_SIZE);
+  key.size = get_unscaled_size (fcfontmap, context, desc);
+
+  if (PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->context_key_get)
+    key.context_key = (gpointer)PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->context_key_get (fcfontmap, context);
+  else
+    key.context_key = NULL;
   
-  fontset_hash = pango_fc_get_fontset_hash (fcfontmap, language);
-  patterns = g_hash_table_lookup (fontset_hash, desc);
+  patterns = g_hash_table_lookup (priv->fontset_hash, &key);
 
   if (patterns == NULL)
     {
-      pattern = pango_fc_make_pattern (desc);
-      if (language)
-	FcPatternAddString (pattern, FC_LANG, (FcChar8 *) pango_language_to_string (language));
-
-      pango_fc_default_substitute (fcfontmap, pattern);
+      pattern = pango_fc_make_pattern (desc, language,
+				       get_scaled_size (&key),
+				       pango_fc_font_map_get_resolution (fcfontmap, context));
       
-      font_patterns = FcFontSort (NULL, pattern, FcTrue, 0, &res);
+      pango_fc_default_substitute (fcfontmap, context, pattern);
+
+      font_patterns = FcFontSort (NULL, pattern, FcTrue, NULL, &res);
 
       if (!font_patterns)
 	{
@@ -828,31 +1131,58 @@ pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
       
       FcFontSetSortDestroy (font_patterns);
 
-      g_hash_table_insert (fontset_hash,
-			   pango_font_description_copy (desc),
+      patterns->key = fontset_hash_key_copy (&key);
+      g_hash_table_insert (priv->fontset_hash,
+			   patterns->key,
 			   patterns);
     }
 
+  if ((!patterns->cache_link ||
+       patterns->cache_link != priv->fontset_cache->head))
+    pango_fc_font_map_cache_fontset (fcfontmap, patterns);
+
+  pango_font_description_free (key.desc);
+
   return patterns;
 }
 
+static gboolean
+get_first_font (PangoFontset  *fontset,
+		PangoFont     *font,
+		gpointer       data)
+{
+  *(PangoFont **)data = font;
+  
+  return TRUE;
+}
+
 static PangoFont *
 pango_fc_font_map_load_font (PangoFontMap               *fontmap,
 			     PangoContext               *context,
 			     const PangoFontDescription *description)
 {
-  PangoFcPatternSet *patterns = pango_fc_font_map_get_patterns (fontmap, context, description, NULL);
-  if (!patterns)
-    return NULL;
+  PangoLanguage *language;
+  PangoFontset *fontset;
+  PangoFont *font = NULL;
 
-  if (patterns->n_patterns > 0)
+  if (context)
+    language = pango_context_get_language (context);
+  else
+    language = NULL;
+
+  fontset = pango_font_map_load_fontset (fontmap, context, description, language);
+
+  if (fontset)
     {
-      return pango_fc_font_map_new_font (fontmap,
-					 context ? pango_context_get_matrix (context) : NULL,
-					 patterns->patterns[0]);
+      pango_fontset_foreach (fontset, get_first_font, &font);
+
+      if (font)
+	g_object_ref (font);
+  
+      g_object_unref (fontset);
     }
   
-  return NULL;
+  return font;
 }
 
 static void
@@ -861,8 +1191,7 @@ pango_fc_pattern_set_free (PangoFcPatternSet *patterns)
   int i;
 
   if (patterns->fontset)
-    g_object_remove_weak_pointer (G_OBJECT (patterns->fontset),
-				  (gpointer *)&patterns->fontset);
+    g_object_unref (patterns->fontset);
   
   for (i = 0; i < patterns->n_patterns; i++)
     FcPatternDestroy (patterns->patterns[i]);
@@ -896,10 +1225,9 @@ pango_fc_font_map_cache_fontset (PangoFcFontMap    *fcfontmap,
 	{
 	  PangoFcPatternSet *tmp_patterns = g_queue_pop_tail (cache);
 	  tmp_patterns->cache_link = NULL;
-	  g_object_unref (tmp_patterns->fontset);
+	  g_hash_table_remove (priv->fontset_hash, tmp_patterns->key);
 	}
 	
-      g_object_ref (patterns->fontset);
       patterns->cache_link = g_list_prepend (NULL, patterns);
     }
 
@@ -913,59 +1241,43 @@ pango_fc_font_map_load_fontset (PangoFontMap                 *fontmap,
 				PangoLanguage                *language)
 {
   PangoFcPatternSet *patterns = pango_fc_font_map_get_patterns (fontmap, context, desc, language);
-  PangoFcFontMap *fcfontmap = PANGO_FC_FONT_MAP (fontmap);
-  PangoFcFontMapPrivate *priv = fcfontmap->priv;
   PangoFontset *result;
-  PangoMatrix *matrix;
   int i;
   
   if (!patterns)
     return NULL;
 
-  if (context)
-    matrix = pango_context_get_matrix (context);
-  else
-    matrix = NULL;
-
-  /* We never cache fontsets when a transformation is in place
-   */
-  if (!patterns->fontset || matrix)
+  if (!patterns->fontset)
     {
       PangoFontsetSimple *simple;
       simple = pango_fontset_simple_new (language);
       result = PANGO_FONTSET (simple);
-      
+
       for (i = 0; i < patterns->n_patterns; i++)
 	{
 	  PangoFont *font;
 
-	  font = pango_fc_font_map_new_font (fontmap, matrix, patterns->patterns[i]);
+	  font = pango_fc_font_map_new_font (fontmap, context, desc,
+					     patterns->patterns[i]);
 	  if (font)
 	    pango_fontset_simple_append (simple, font);
 	}
 
-      if (!matrix)
-	{
-	  patterns->fontset = PANGO_FONTSET (simple);
-	  g_object_add_weak_pointer (G_OBJECT (patterns->fontset),
-				     (gpointer *)&patterns->fontset);
-	}
+      patterns->fontset = PANGO_FONTSET (simple);
     }
-  else
-    result = g_object_ref (patterns->fontset);
   
-  if (!matrix &&
-      (!patterns->cache_link ||
-       patterns->cache_link != priv->fontset_cache->head))
-    pango_fc_font_map_cache_fontset (fcfontmap, patterns);
-
+  result = g_object_ref (patterns->fontset);
+  
   return result;
 }
 
 static void
-uncache_patterns (PangoFcPatternSet *patterns)
+uncache_patterns (PangoFcPatternSet *patterns,
+		  PangoFcFontMap    *fcfontmap)
 {
-  g_object_unref (patterns->fontset);
+  PangoFcFontMapPrivate *priv = fcfontmap->priv;
+  
+  g_hash_table_remove (priv->fontset_hash, patterns->key);
 }
 
 static void
@@ -974,7 +1286,7 @@ pango_fc_font_map_clear_fontset_cache (PangoFcFontMap *fcfontmap)
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
   GQueue *cache = priv->fontset_cache;
   
-  g_list_foreach (cache->head, (GFunc)uncache_patterns, NULL);
+  g_list_foreach (cache->head, (GFunc)uncache_patterns, fcfontmap);
   g_list_free (cache->head);
   cache->head = NULL;
   cache->tail = NULL;
@@ -997,11 +1309,9 @@ pango_fc_font_map_clear_fontset_cache (PangoFcFontMap *fcfontmap)
 void
 pango_fc_font_map_cache_clear (PangoFcFontMap *fcfontmap)
 {
-  /* Clear the fontset cache first, since any entries
-   * in the fontset_cache must also be in the pattern cache.
-   */
+  fcfontmap->priv->dpi = -1;
+  
   pango_fc_font_map_clear_fontset_cache (fcfontmap);
-  pango_fc_clear_pattern_hashes (fcfontmap);
 }
 
 static void
@@ -1170,28 +1480,51 @@ pango_fc_font_map_shutdown (PangoFcFontMap *fcfontmap)
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
 
   pango_fc_font_map_cache_clear (fcfontmap);
+  g_hash_table_destroy (priv->fontset_hash);
+  priv->fontset_hash = NULL;
   
-  g_hash_table_foreach (priv->fonts, (GHFunc)cleanup_font, NULL);
-  g_hash_table_destroy (priv->fonts);
-  priv->fonts = NULL;
+  g_hash_table_foreach (priv->font_hash, (GHFunc)cleanup_font, NULL);
+  g_hash_table_destroy (priv->font_hash);
+  priv->font_hash = NULL;
   priv->closed = TRUE;
 }
 
 static PangoWeight
 pango_fc_convert_weight_to_pango (int fc_weight)
 {
-  if (fc_weight < FC_WEIGHT_LIGHT)
+#ifdef FC_WEIGHT_ULTRABOLD
+  /* fontconfig 2.1 only had light/medium/demibold/bold/black */
+  if (fc_weight < (FC_WEIGHT_ULTRALIGHT + FC_WEIGHT_LIGHT) / 2)
     return PANGO_WEIGHT_ULTRALIGHT;
-  else if (fc_weight < (FC_WEIGHT_LIGHT + FC_WEIGHT_MEDIUM) / 2)
+  else if (fc_weight < (FC_WEIGHT_LIGHT + FC_WEIGHT_REGULAR) / 2)
+    return PANGO_WEIGHT_LIGHT;
+  else if (fc_weight < (FC_WEIGHT_REGULAR + FC_WEIGHT_DEMIBOLD) / 2)
+    return PANGO_WEIGHT_NORMAL;
+  /* We group the 500/MEDIUM weight with normal to reduce confusion
+   *
+   * else if (fc_weight < (FC_WEIGHT_MEDIUM + FC_WEIGHT_DEMIBOLD) / 2)
+   *  return 500;
+   */
+  else if (fc_weight < (FC_WEIGHT_DEMIBOLD + FC_WEIGHT_BOLD) / 2)
+    return PANGO_WEIGHT_SEMIBOLD;
+  else if (fc_weight < (FC_WEIGHT_BOLD + FC_WEIGHT_ULTRABOLD) / 2)
+    return PANGO_WEIGHT_BOLD;
+  else if (fc_weight < (FC_WEIGHT_ULTRABOLD + FC_WEIGHT_BLACK) / 2)
+    return PANGO_WEIGHT_ULTRABOLD;
+  else
+    return PANGO_WEIGHT_HEAVY;
+#else  /* fontconfig < 2.2 */
+  if (fc_weight < (FC_WEIGHT_LIGHT + FC_WEIGHT_MEDIUM) / 2)
     return PANGO_WEIGHT_LIGHT;
   else if (fc_weight < (FC_WEIGHT_MEDIUM + FC_WEIGHT_DEMIBOLD) / 2)
     return PANGO_WEIGHT_NORMAL;
   else if (fc_weight < (FC_WEIGHT_DEMIBOLD + FC_WEIGHT_BOLD) / 2)
-    return 600;
+    return PANGO_WEIGHT_SEMIBOLD;
   else if (fc_weight < (FC_WEIGHT_BOLD + FC_WEIGHT_BLACK) / 2)
     return PANGO_WEIGHT_BOLD;
   else
     return PANGO_WEIGHT_ULTRABOLD;
+#endif  
 }
 
 static PangoStyle
@@ -1245,6 +1578,7 @@ pango_fc_convert_width_to_pango (int fc_stretch)
  * @pattern: a #FcPattern
  * @include_size: if %TRUE, the pattern will include the size from
  *   the @pattern; otherwise the resulting pattern will be unsized.
+ *   (only %FC_SIZE is examined, not %FC_PIXEL_SIZE)
  * 
  * Creates a #PangoFontDescription that matches the specified
  * Fontconfig pattern as closely as possible. Many possible Fontconfig
@@ -1423,17 +1757,8 @@ pango_fc_face_list_sizes (PangoFontFace  *face,
           if (FcPatternGetDouble (fontset->fonts[i], FC_PIXEL_SIZE, 0, &size) == FcResultMatch)
             {
               if (dpi < 0)
-                {
-                  FcPattern *tmp = FcPatternDuplicate (fontset->fonts[i]);
-                  pango_fc_default_substitute (fcface->family->fontmap, tmp);
-                  if (FcPatternGetDouble (tmp, FC_DPI, 0, &dpi) != FcResultMatch)
-                    {
-                      g_warning ("Error getting DPI from fontconfig, using 72.0");
-                      dpi = 72.0;
-                    }
-                  FcPatternDestroy (tmp);
-                }
-
+		dpi = pango_fc_font_map_get_resolution (fcface->family->fontmap, NULL);
+	      
               size_i = (int) (PANGO_SCALE * size * 72.0 / dpi);
               g_array_append_val (size_array, size_i);
             }
